<!DOCTYPE html>
<html lang="en">

<head>
	<meta charset="utf-8">
	<title>{{ s_doctitle }}</title>
	<meta http-equiv="X-UA-Compatible" content="IE=edge">
	<meta name="viewport" content="width=device-width, initial-scale=0.8">
	<meta name="theme-color" content="#{{ tcolor }}">
<style>:root{--font-main:sans-serif;--font-mono:monospace}
html,body,a{margin:0;padding:0;border:none;color:#ccc;background:none;font-family:sans-serif;font-family:var(--font-main),sans-serif}
pre{font-family:monospace,monospace;font-family:var(--font-mono),monospace}
html{touch-action:manipulation;background:#333}
#box{padding:.5em 1em;background:#2c2c2c}
a{color:#fc5}</style>
{{ html_head }}
</head>

<body>
	<div id="box">

		{%- if h1 %}
		<h1>{{ h1 }}</h1>
		{%- endif %}

		{%- if h2 %}
		<h2>{{ h2 }}</h2>
		{%- endif %}

		{%- if p %}
		<p>{{ p }}</p>
		{%- endif %}

		{%- if pre %}
		<pre>{{ pre }}</pre>
		{%- endif %}

		{%- if html %}
		{{ html }}
		{%- endif %}

		{%- if click %}
		<script>document.getElementsByTagName("a")[0].click()</script>
		{%- endif %}
	</div>

	{%- if redir %}
	<script>
		setTimeout(function() {
			location.replace("{{ redir }}");
		}, 800);
	</script>
	{%- endif %}
	{%- if js %}
	<script src="{{ js }}_={{ ts }}"></script>
	{%- endif %}
</body>

</html>

